// CGlmModel.cpp: implementation of the CGlmModel class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Kurushi.h"
#include "GlmModel.h"
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

/* defines */
#define T(x) triangles[(x)]


/* enums */
enum { X = 0, Y, Z, W };			/* elements of a vertex */


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CGlmModel::CGlmModel()
{
  pathname      = NULL;
  mtllibname    = NULL;
  numvertices   = 0;
  vertices      = NULL;
  numnormals    = 0;
  normals       = NULL;
  numtexcoords  = 0;
  texcoords     = NULL;
  numfacetnorms = 0;
  facetnorms    = NULL;
  numtriangles  = 0;
  triangles     = NULL;
  nummaterials  = 0;
  materials     = NULL;
  numgroups     = 0;
  groups        = NULL;
  position[0]   = 0.0;
  position[1]   = 0.0;
  position[2]   = 0.0;

}

CGlmModel::~CGlmModel()
{
  GLMgroup* group;
  GLuint i;

  if (this->pathname)   free(this->pathname);
  if (this->mtllibname) free(this->mtllibname);
  if (this->vertices)   free(this->vertices);
  if (this->normals)    free(this->normals);
  if (this->texcoords)  free(this->texcoords);
  if (this->facetnorms) free(this->facetnorms);
  if (this->triangles)  free(this->triangles);
  if (this->materials) {
    for (i = 0; i < this->nummaterials; i++)
      free(this->materials[i].name);
  }
  free(this->materials);
  while(this->groups) {
    group = this->groups;
    this->groups = this->groups->next;
    free(group->name);
    free(group->triangles);
    free(group);
  }
}


/* _glmDot: compute the dot product of two vectors
 *
 * u - array of 3 GLfloats (GLfloat u[3])
 * v - array of 3 GLfloats (GLfloat v[3])
 */
GLfloat CGlmModel::_glmDot(GLfloat* u, GLfloat* v)
{
  ASSERT(u);
  ASSERT(v);

  /* compute the dot product */
  return u[X] * v[X] + u[Y] * v[Y] + u[Z] * v[Z];
}

/* _glmCross: compute the cross product of two vectors
 *
 * u - array of 3 GLfloats (GLfloat u[3])
 * v - array of 3 GLfloats (GLfloat v[3])
 * n - array of 3 GLfloats (GLfloat n[3]) to return the cross product in
 */
GLvoid CGlmModel::_glmCross(GLfloat* u, GLfloat* v, GLfloat* n)
{
  ASSERT(u);
  ASSERT(v);
  ASSERT(n);

  /* compute the cross product (u x v for right-handed [ccw]) */
  n[X] = u[Y] * v[Z] - u[Z] * v[Y];
  n[Y] = u[Z] * v[X] - u[X] * v[Z];
  n[Z] = u[X] * v[Y] - u[Y] * v[X];
}

/* _glmNormalize: normalize a vector
 *
 * n - array of 3 GLfloats (GLfloat n[3]) to be normalized
 */
GLvoid CGlmModel::_glmNormalize(GLfloat* n)
{
  GLfloat l;

  ASSERT(n);

  /* normalize */
  l = (GLfloat)sqrt(n[X] * n[X] + n[Y] * n[Y] + n[Z] * n[Z]);
  n[0] /= l;
  n[1] /= l;
  n[2] /= l;
}

/* _glmEqual: compares two vectors and returns GL_TRUE if they are
 * equal (within a certain threshold) or GL_FALSE if not. An epsilon
 * that works fairly well is 0.000001.
 *
 * u - array of 3 GLfloats (GLfloat u[3])
 * v - array of 3 GLfloats (GLfloat v[3]) 
 */
 GLboolean CGlmModel::_glmEqual(GLfloat* u, GLfloat* v, GLfloat epsilon)
{
  if (_glmAbs(u[0] - v[0]) < epsilon &&
      _glmAbs(u[1] - v[1]) < epsilon &&
      _glmAbs(u[2] - v[2]) < epsilon) 
  {
    return GL_TRUE;
  }
  return GL_FALSE;
}

/* _glmWeldVectors: eliminate (weld) vectors that are within an
 * epsilon of each other.
 *
 * vectors    - array of GLfloat[3]'s to be welded
 * numvectors - number of GLfloat[3]'s in vectors
 * epsilon    - maximum difference between vectors 
 *
 */
GLfloat* CGlmModel::_glmWeldVectors(GLfloat* vectors, GLuint* numvectors, GLfloat epsilon)
{
  GLfloat* copies;
  GLuint   copied;
  GLuint   i, j;

  copies = (GLfloat*)malloc(sizeof(GLfloat) * 3 * (*numvectors + 1));
  memcpy(copies, vectors, (sizeof(GLfloat) * 3 * (*numvectors + 1)));

  copied = 1;
  for (i = 1; i <= *numvectors; i++) {
    for (j = 1; j <= copied; j++) {
      if (_glmEqual(&vectors[3 * i], &copies[3 * j], epsilon)) {
	goto duplicate;
      }
    }

    /* must not be any duplicates -- add to the copies array */
    copies[3 * copied + 0] = vectors[3 * i + 0];
    copies[3 * copied + 1] = vectors[3 * i + 1];
    copies[3 * copied + 2] = vectors[3 * i + 2];
    j = copied;				/* pass this along for below */
    copied++;

  duplicate:
    /* set the first component of this vector to point at the correct
       index into the new copies array */
    vectors[3 * i + 0] = (GLfloat)j;
  }

  *numvectors = copied-1;
  return copies;
}

/* _glmFindGroup: Find a group in the model
 */
GLMgroup* CGlmModel::_glmFindGroup(char* name)
{
  GLMgroup* group;

  

  group = this->groups;
  while(group) {
    if (!strcmp(name, group->name))
      break;
    group = group->next;
  }

  return group;
}

/* _glmAddGroup: Add a group to the model
 */
GLMgroup* CGlmModel::_glmAddGroup(char* name)
{
  GLMgroup* group;

  group = _glmFindGroup(name);
  if (!group) {
    group = (GLMgroup*)malloc(sizeof(GLMgroup));
    group->name = strdup(name);
    group->material = 0;
    group->numtriangles = 0;
    group->triangles = NULL;
    group->next = this->groups;
    this->groups = group;
    this->numgroups++;
  }

  return group;
}

/* _glmFindGroup: Find a material in the model
 */
GLuint CGlmModel::_glmFindMaterial(char* name)
{
  GLuint i;

  for (i = 0; i < this->nummaterials; i++) {
    if (!strcmp(this->materials[i].name, name))
      goto found;
  }

  /* didn't find the name, so set it as the default material */
  printf("_glmFindMaterial():  can't find material \"%s\".\n", name);
  i = 0;

found:
  return i;
}


/* _glmDirName: return the directory given a path
 *
 * path - filesystem path
 *
 * The return value should be free'd.
 */
char* CGlmModel::_glmDirName(char* path)
{
  char* dir;
  char* s;

  dir = strdup(path);

  s = strrchr(dir, '/');
  if (s)
    s[1] = '\0';
  else
    dir[0] = '\0';

  return dir;
}


/* _glmReadMTL: read a wavefront material library file
 *
 * model - properly initialized GLMmodel structure
 * name  - name of the material library
 */
GLvoid CGlmModel::_glmReadMTL(char* name)
{
  FILE* file;
  char* dir;
  char* filename;
  char  buf[128];
  GLuint nummaterials, i;

  dir = _glmDirName(this->pathname);
  filename = (char*)malloc(sizeof(char) * (strlen(dir) + strlen(name) + 1));
  strcpy(filename, dir);
  strcat(filename, name);
  free(dir);

  /* open the file */
  file = fopen(filename, "r");
  if (!file) {
    TRACE("_glmReadMTL() failed: can't open material file \"%s\".\n", filename);
    exit(1);
  }
  free(filename);

  /* count the number of materials in the file */
  nummaterials = 1;
  while(fscanf(file, "%s", buf) != EOF) {
    switch(buf[0]) {
    case '#':				/* comment */
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    case 'n':				/* newmtl */
      fgets(buf, sizeof(buf), file);
      nummaterials++;
      sscanf(buf, "%s %s", buf, buf);
      break;
    default:
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    }
  }

  rewind(file);

  /* allocate memory for the materials */
  this->materials = (GLMmaterial*)malloc(sizeof(GLMmaterial) * nummaterials);
  this->nummaterials = nummaterials;

  /* set the default material */
  for (i = 0; i < nummaterials; i++) {
    this->materials[i].name = NULL;
    this->materials[i].shininess = 0;
    this->materials[i].diffuse[0] = 0.8f;
    this->materials[i].diffuse[1] = 0.8f;
    this->materials[i].diffuse[2] = 0.8f;
    this->materials[i].diffuse[3] = 1.0f;
    this->materials[i].ambient[0] = 0.2f;
    this->materials[i].ambient[1] = 0.2f;
    this->materials[i].ambient[2] = 0.2f;
    this->materials[i].ambient[3] = 1.0f;
    this->materials[i].specular[0] = 0.0f;
    this->materials[i].specular[1] = 0.0f;
    this->materials[i].specular[2] = 0.0f;
    this->materials[i].specular[3] = 1.0f;
  }
  this->materials[0].name = strdup("default");

  /* now, read in the data */
  nummaterials = 0;
  while(fscanf(file, "%s", buf) != EOF) {
    switch(buf[0]) {
    case '#':				/* comment */
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    case 'n':				/* newmtl */
      fgets(buf, sizeof(buf), file);
      sscanf(buf, "%s %s", buf, buf);
      nummaterials++;
      this->materials[nummaterials].name = strdup(buf);
      break;
    case 'N':
      fscanf(file, "%f", &this->materials[nummaterials].shininess);
      /* wavefront shininess is from [0, 1000], so scale for OpenGL */
      this->materials[nummaterials].shininess /= 1000.0;
      this->materials[nummaterials].shininess *= 128.0;
      break;
    case 'K':
      switch(buf[1]) {
      case 'd':
	fscanf(file, "%f %f %f",
	       &this->materials[nummaterials].diffuse[0],
	       &this->materials[nummaterials].diffuse[1],
	       &this->materials[nummaterials].diffuse[2]);
	break;
      case 's':
	fscanf(file, "%f %f %f",
	       &this->materials[nummaterials].specular[0],
	       &this->materials[nummaterials].specular[1],
	       &this->materials[nummaterials].specular[2]);
	break;
      case 'a':
	fscanf(file, "%f %f %f",
	       &this->materials[nummaterials].ambient[0],
	       &this->materials[nummaterials].ambient[1],
	       &this->materials[nummaterials].ambient[2]);
	break;
      default:
	/* eat up rest of line */
	fgets(buf, sizeof(buf), file);
	break;
      }
      break;
    default:
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    }
  }
}

/* _glmWriteMTL: write a wavefront material library file
 *
 * model      - properly initialized GLMmodel structure
 * modelpath  - pathname of the model being written
 * mtllibname - name of the material library to be written
 */
GLvoid CGlmModel::_glmWriteMTL(char* modelpath, char* mtllibname)
{
  FILE* file;
  char* dir;
  char* filename;
  GLMmaterial* material;
  GLuint i;

  dir = _glmDirName(modelpath);
  filename = (char*)malloc(sizeof(char) * (strlen(dir) + strlen(mtllibname)));
  strcpy(filename, dir);
  strcat(filename, mtllibname);
  free(dir);

  /* open the file */
  file = fopen(filename, "w");
  if (!file) {
    TRACE("_glmWriteMTL() failed: can't open file \"%s\".\n", filename); exit(1);}
  free(filename);

  /* spit out a header */
  fprintf(file, "#  \n");
  fprintf(file, "#  Wavefront MTL generated by GLM library\n");
  fprintf(file, "#  \n");
  fprintf(file, "#  GLM library copyright (C) 1997 by Nate Robins\n");
  fprintf(file, "#  email: ndr@pobox.com\n");
  fprintf(file, "#  www:   http://www.pobox.com/~ndr\n");
  fprintf(file, "#  \n\n");

  for (i = 0; i < this->nummaterials; i++) {
    material = &this->materials[i];
    fprintf(file, "newmtl %s\n", material->name);
    fprintf(file, "Ka %f %f %f\n", 
	    material->ambient[0], material->ambient[1], material->ambient[2]);
    fprintf(file, "Kd %f %f %f\n", 
	    material->diffuse[0], material->diffuse[1], material->diffuse[2]);
    fprintf(file, "Ks %f %f %f\n", 
	    material->specular[0],material->specular[1],material->specular[2]);
    fprintf(file, "Ns %f\n", material->shininess);
    fprintf(file, "\n");
  }
}


/* _glmFirstPass: first pass at a Wavefront OBJ file that gets all the
 * statistics of the model (such as #vertices, #normals, etc)
 *
 * model - properly initialized GLMmodel structure
 * file  - (fopen'd) file descriptor 
 */
GLvoid CGlmModel::_glmFirstPass(FILE* file) 
{
  GLuint    numvertices;		/* number of vertices in model */
  GLuint    numnormals;			/* number of normals in model */
  GLuint    numtexcoords;		/* number of texcoords in model */
  GLuint    numtriangles;		/* number of triangles in model */
  GLMgroup* group;			/* current group */
  unsigned  v, n, t;
  char      buf[128];

  /* make a default group */
  group = _glmAddGroup("default");

  numvertices = numnormals = numtexcoords = numtriangles = 0;
  while(fscanf(file, "%s", buf) != EOF) {
    switch(buf[0]) {
    case '#':				/* comment */
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    case 'v':				/* v, vn, vt */
      switch(buf[1]) {
      case '\0':			/* vertex */
	/* eat up rest of line */
	fgets(buf, sizeof(buf), file);
	numvertices++;
	break;
      case 'n':				/* normal */
	/* eat up rest of line */
	fgets(buf, sizeof(buf), file);
	numnormals++;
	break;
      case 't':				/* texcoord */
	/* eat up rest of line */
	fgets(buf, sizeof(buf), file);
	numtexcoords++;
	break;
      default:
	printf("_glmFirstPass(): Unknown token \"%s\".\n", buf);
	exit(1);
	break;
      }
      break;
    case 'm':
      fgets(buf, sizeof(buf), file);
      sscanf(buf, "%s %s", buf, buf);
      this->mtllibname = strdup(buf);
      _glmReadMTL(buf);
      break;
    case 'u':
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    case 'g':				/* group */
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      sscanf(buf, "%s", buf);
      group = _glmAddGroup(buf);
      break;
    case 'f':				/* face */
      v = n = t = 0;
      fscanf(file, "%s", buf);
      /* can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d */
      if (strstr(buf, "//")) {
	/* v//n */
	sscanf(buf, "%d//%d", &v, &n);
	fscanf(file, "%d//%d", &v, &n);
	fscanf(file, "%d//%d", &v, &n);
	numtriangles++;
	group->numtriangles++;
	while(fscanf(file, "%d//%d", &v, &n) > 0) {
	  numtriangles++;
	  group->numtriangles++;
	}
      } else if (sscanf(buf, "%d/%d/%d", &v, &t, &n) == 3) {
	/* v/t/n */
	fscanf(file, "%d/%d/%d", &v, &t, &n);
	fscanf(file, "%d/%d/%d", &v, &t, &n);
	numtriangles++;
	group->numtriangles++;
	while(fscanf(file, "%d/%d/%d", &v, &t, &n) > 0) {
	  numtriangles++;
	  group->numtriangles++;
	}
      } else if (sscanf(buf, "%d/%d", &v, &t) == 2) {
	/* v/t */
	fscanf(file, "%d/%d", &v, &t);
	fscanf(file, "%d/%d", &v, &t);
	numtriangles++;
	group->numtriangles++;
	while(fscanf(file, "%d/%d", &v, &t) > 0) {
	  numtriangles++;
	  group->numtriangles++;
	}
      } else {
	/* v */
	fscanf(file, "%d", &v);
	fscanf(file, "%d", &v);
	numtriangles++;
	group->numtriangles++;
	while(fscanf(file, "%d", &v) > 0) {
	  numtriangles++;
	  group->numtriangles++;
	}
      }
      break;

    default:
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    }
  }

  /* set the stats in the model structure */
  this->numvertices  = numvertices;
  this->numnormals   = numnormals;
  this->numtexcoords = numtexcoords;
  this->numtriangles = numtriangles;

  /* allocate memory for the triangles in each group */
  group = this->groups;
  while(group) {
    group->triangles = (GLuint*)malloc(sizeof(GLuint) * group->numtriangles);
    group->numtriangles = 0;
    group = group->next;
  }
}

/* _glmSecondPass: second pass at a Wavefront OBJ file that gets all
 * the data.
 *
 * model - properly initialized GLMmodel structure
 * file  - (fopen'd) file descriptor 
 */
GLvoid CGlmModel::_glmSecondPass(FILE* file) 
{
  GLuint    numvertices;		/* number of vertices in model */
  GLuint    numnormals;			/* number of normals in model */
  GLuint    numtexcoords;		/* number of texcoords in model */
  GLuint    numtriangles;		/* number of triangles in model */
  GLfloat*  vertices;			/* array of vertices  */
  GLfloat*  normals;			/* array of normals */
  GLfloat*  texcoords;			/* array of texture coordinates */
  GLMgroup* group;			/* current group pointer */
  GLuint    material;			/* current material */
  GLuint    v, n, t;
  char      buf[128];

  /* set the pointer shortcuts */
  vertices     = this->vertices;
  normals      = this->normals;
  texcoords    = this->texcoords;
  group        = this->groups;

  /* on the second pass through the file, read all the data into the
     allocated arrays */
  numvertices = numnormals = numtexcoords = 1;
  numtriangles = 0;
  material = 0;
  while(fscanf(file, "%s", buf) != EOF) {
    switch(buf[0]) {
    case '#':				/* comment */
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    case 'v':				/* v, vn, vt */
      switch(buf[1]) {
      case '\0':			/* vertex */
	fscanf(file, "%f %f %f", 
	       &vertices[3 * numvertices + X], 
	       &vertices[3 * numvertices + Y], 
	       &vertices[3 * numvertices + Z]);
	numvertices++;
	break;
      case 'n':				/* normal */
	fscanf(file, "%f %f %f", 
	       &normals[3 * numnormals + X],
	       &normals[3 * numnormals + Y], 
	       &normals[3 * numnormals + Z]);
	numnormals++;
	break;
      case 't':				/* texcoord */
	fscanf(file, "%f %f", 
	       &texcoords[2 * numtexcoords + X],
	       &texcoords[2 * numtexcoords + Y]);
	numtexcoords++;
	break;
      }
      break;
    case 'u':
      fgets(buf, sizeof(buf), file);
      sscanf(buf, "%s %s", buf, buf);
      group->material = material = _glmFindMaterial(buf);
      break;
    case 'g':				/* group */
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      sscanf(buf, "%s", buf);
      group = _glmFindGroup(buf);
      group->material = material;
      break;
    case 'f':				/* face */
      v = n = t = 0;
      fscanf(file, "%s", buf);
      /* can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d */
      if (strstr(buf, "//")) {
	/* v//n */
	sscanf(buf, "%d//%d", &v, &n);
	T(numtriangles).vindices[0] = v;
	T(numtriangles).nindices[0] = n;
	fscanf(file, "%d//%d", &v, &n);
	T(numtriangles).vindices[1] = v;
	T(numtriangles).nindices[1] = n;
	fscanf(file, "%d//%d", &v, &n);
	T(numtriangles).vindices[2] = v;
	T(numtriangles).nindices[2] = n;
	group->triangles[group->numtriangles++] = numtriangles;
	numtriangles++;
	while(fscanf(file, "%d//%d", &v, &n) > 0) {
	  T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
	  T(numtriangles).nindices[0] = T(numtriangles-1).nindices[0];
	  T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
	  T(numtriangles).nindices[1] = T(numtriangles-1).nindices[2];
	  T(numtriangles).vindices[2] = v;
	  T(numtriangles).nindices[2] = n;
	  group->triangles[group->numtriangles++] = numtriangles;
	  numtriangles++;
	}
      } else if (sscanf(buf, "%d/%d/%d", &v, &t, &n) == 3) {
	/* v/t/n */
	T(numtriangles).vindices[0] = v;
	T(numtriangles).tindices[0] = t;
	T(numtriangles).nindices[0] = n;
	fscanf(file, "%d/%d/%d", &v, &t, &n);
	T(numtriangles).vindices[1] = v;
	T(numtriangles).tindices[1] = t;
	T(numtriangles).nindices[1] = n;
	fscanf(file, "%d/%d/%d", &v, &t, &n);
	T(numtriangles).vindices[2] = v;
	T(numtriangles).tindices[2] = t;
	T(numtriangles).nindices[2] = n;
	group->triangles[group->numtriangles++] = numtriangles;
	numtriangles++;
	while(fscanf(file, "%d/%d/%d", &v, &t, &n) > 0) {
	  T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
	  T(numtriangles).tindices[0] = T(numtriangles-1).tindices[0];
	  T(numtriangles).nindices[0] = T(numtriangles-1).nindices[0];
	  T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
	  T(numtriangles).tindices[1] = T(numtriangles-1).tindices[2];
	  T(numtriangles).nindices[1] = T(numtriangles-1).nindices[2];
	  T(numtriangles).vindices[2] = v;
	  T(numtriangles).tindices[2] = t;
	  T(numtriangles).nindices[2] = n;
	  group->triangles[group->numtriangles++] = numtriangles;
	  numtriangles++;
	}
      } else if (sscanf(buf, "%d/%d", &v, &t) == 2) {
	/* v/t */
	T(numtriangles).vindices[0] = v;
	T(numtriangles).tindices[0] = t;
	fscanf(file, "%d/%d", &v, &t);
	T(numtriangles).vindices[1] = v;
	T(numtriangles).tindices[1] = t;
	fscanf(file, "%d/%d", &v, &t);
	T(numtriangles).vindices[2] = v;
	T(numtriangles).tindices[2] = t;
	group->triangles[group->numtriangles++] = numtriangles;
	numtriangles++;
	while(fscanf(file, "%d/%d", &v, &t) > 0) {
	  T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
	  T(numtriangles).tindices[0] = T(numtriangles-1).tindices[0];
	  T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
	  T(numtriangles).tindices[1] = T(numtriangles-1).tindices[2];
	  T(numtriangles).vindices[2] = v;
	  T(numtriangles).tindices[2] = t;
	  group->triangles[group->numtriangles++] = numtriangles;
	  numtriangles++;
	}
      } else {
	/* v */
	sscanf(buf, "%d", &v);
	T(numtriangles).vindices[0] = v;
	fscanf(file, "%d", &v);
	T(numtriangles).vindices[1] = v;
	fscanf(file, "%d", &v);
	T(numtriangles).vindices[2] = v;
	group->triangles[group->numtriangles++] = numtriangles;
	numtriangles++;
	while(fscanf(file, "%d", &v) > 0) {
	  T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
	  T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
	  T(numtriangles).vindices[2] = v;
	  group->triangles[group->numtriangles++] = numtriangles;
	  numtriangles++;
	}
      }
      break;

    default:
      /* eat up rest of line */
      fgets(buf, sizeof(buf), file);
      break;
    }
  }

#if 0
  /* announce the memory requirements */
  printf(" Memory: %d bytes\n",
	 numvertices  * 3*sizeof(GLfloat) +
	 numnormals   * 3*sizeof(GLfloat) * (numnormals ? 1 : 0) +
	 numtexcoords * 3*sizeof(GLfloat) * (numtexcoords ? 1 : 0) +
	 numtriangles * sizeof(GLMtriangle));
#endif
}




/* public functions */

/* glmUnitize: "unitize" a model by translating it to the origin and
 * scaling it to fit in a unit cube around the origin.  Returns the
 * scalefactor used.
 *
 * model - properly initialized GLMmodel structure 
 */
GLfloat CGlmModel::glmUnitize(GLfloat u)
{
  GLuint  i;
  GLfloat maxx, minx, maxy, miny, maxz, minz;
  GLfloat cx, cy, cz, w, h, d;
  GLfloat scale;

  
  assert(this->vertices);

  /* get the max/mins */
  maxx = minx = this->vertices[3 + X];
  maxy = miny = this->vertices[3 + Y];
  maxz = minz = this->vertices[3 + Z];
  for (i = 1; i <= this->numvertices; i++) {
    if (maxx < this->vertices[3 * i + X])
      maxx = this->vertices[3 * i + X];
    if (minx > this->vertices[3 * i + X])
      minx = this->vertices[3 * i + X];

    if (maxy < this->vertices[3 * i + Y])
      maxy = this->vertices[3 * i + Y];
    if (miny > this->vertices[3 * i + Y])
      miny = this->vertices[3 * i + Y];

    if (maxz < this->vertices[3 * i + Z])
      maxz = this->vertices[3 * i + Z];
    if (minz > this->vertices[3 * i + Z])
      minz = this->vertices[3 * i + Z];
  }

  /* calculate model width, height, and depth */
  w = _glmAbs(maxx) + _glmAbs(minx);
  h = _glmAbs(maxy) + _glmAbs(miny);
  d = _glmAbs(maxz) + _glmAbs(minz);

  /* calculate center of the model */
  cx = (maxx + minx) / 2.0f;
  cy = (maxy + miny) / 2.0f;
  cz = (maxz + minz) / 2.0f;

  /* calculate unitizing scale factor */
  scale = u / _glmMax(_glmMax(w, h), d);

  /* translate around center then scale */
  for (i = 1; i <= this->numvertices; i++) {
    this->vertices[3 * i + X] -= cx;
    this->vertices[3 * i + Y] -= cy;
    this->vertices[3 * i + Z] -= cz;
    this->vertices[3 * i + X] *= scale;
    this->vertices[3 * i + Y] *= scale;
    this->vertices[3 * i + Z] *= scale;
  }

  return scale;
}

/* glmDimensions: Calculates the dimensions (width, height, depth) of
 * a model.
 *
 * model      - initialized GLMmodel structure
 * dimensions - array of 3 GLfloats (GLfloat dimensions[3])
 */
GLvoid CGlmModel::glmDimensions(GLfloat* dimensions)
{
  GLuint i;
  GLfloat maxx, minx, maxy, miny, maxz, minz;

  
  assert(this->vertices);
  assert(dimensions);

  /* get the max/mins */
  maxx = minx = this->vertices[3 + X];
  maxy = miny = this->vertices[3 + Y];
  maxz = minz = this->vertices[3 + Z];
  for (i = 1; i <= this->numvertices; i++) {
    if (maxx < this->vertices[3 * i + X])
      maxx = this->vertices[3 * i + X];
    if (minx > this->vertices[3 * i + X])
      minx = this->vertices[3 * i + X];

    if (maxy < this->vertices[3 * i + Y])
      maxy = this->vertices[3 * i + Y];
    if (miny > this->vertices[3 * i + Y])
      miny = this->vertices[3 * i + Y];

    if (maxz < this->vertices[3 * i + Z])
      maxz = this->vertices[3 * i + Z];
    if (minz > this->vertices[3 * i + Z])
      minz = this->vertices[3 * i + Z];
  }

  /* calculate model width, height, and depth */
  dimensions[X] = _glmAbs(maxx) + _glmAbs(minx);
  dimensions[Y] = _glmAbs(maxy) + _glmAbs(miny);
  dimensions[Z] = _glmAbs(maxz) + _glmAbs(minz);
}

/* glmScale: Scales a model by a given amount.
 * 
 * model - properly initialized GLMmodel structure
 * scale - scalefactor (0.5 = half as large, 2.0 = twice as large)
 */
GLvoid CGlmModel::glmScale(GLfloat scale)
{
  GLuint i;

  for (i = 1; i <= this->numvertices; i++) {
    this->vertices[3 * i + X] *= scale;
    this->vertices[3 * i + Y] *= scale;
    this->vertices[3 * i + Z] *= scale;
  }
}

/* glmReverseWinding: Reverse the polygon winding for all polygons in
 * this model.  Default winding is counter-clockwise.  Also changes
 * the direction of the normals.
 * 
 * model - properly initialized GLMmodel structure 
 */
GLvoid CGlmModel::glmReverseWinding()
{
  GLuint i, swap;

  

  for (i = 0; i < this->numtriangles; i++) {
    swap = T(i).vindices[0];
    T(i).vindices[0] = T(i).vindices[2];
    T(i).vindices[2] = swap;

    if (this->numnormals) {
      swap = T(i).nindices[0];
      T(i).nindices[0] = T(i).nindices[2];
      T(i).nindices[2] = swap;
    }

    if (this->numtexcoords) {
      swap = T(i).tindices[0];
      T(i).tindices[0] = T(i).tindices[2];
      T(i).tindices[2] = swap;
    }
  }

  /* reverse facet normals */
  for (i = 1; i <= this->numfacetnorms; i++) {
    this->facetnorms[3 * i + X] = -this->facetnorms[3 * i + X];
    this->facetnorms[3 * i + Y] = -this->facetnorms[3 * i + Y];
    this->facetnorms[3 * i + Z] = -this->facetnorms[3 * i + Z];
  }

  /* reverse vertex normals */
  for (i = 1; i <= this->numnormals; i++) {
    this->normals[3 * i + X] = -this->normals[3 * i + X];
    this->normals[3 * i + Y] = -this->normals[3 * i + Y];
    this->normals[3 * i + Z] = -this->normals[3 * i + Z];
  }
}

/* glmFacetNormals: Generates facet normals for a model (by taking the
 * cross product of the two vectors derived from the sides of each
 * triangle).  Assumes a counter-clockwise winding.
 *
 * model - initialized GLMmodel structure
 */
GLvoid CGlmModel::glmFacetNormals()
{
  GLuint  i;
  GLfloat u[3];
  GLfloat v[3];
  
  
  assert(this->vertices);

  /* clobber any old facetnormals */
  if (this->facetnorms)
    free(this->facetnorms);

  /* allocate memory for the new facet normals */
  this->numfacetnorms = this->numtriangles;
  this->facetnorms = (GLfloat*)malloc(sizeof(GLfloat) *
				       3 * (this->numfacetnorms + 1));

  for (i = 0; i < this->numtriangles; i++) {
    this->triangles[i].findex = i+1;

    u[X] = this->vertices[3 * T(i).vindices[1] + X] -
           this->vertices[3 * T(i).vindices[0] + X];
    u[Y] = this->vertices[3 * T(i).vindices[1] + Y] -
           this->vertices[3 * T(i).vindices[0] + Y];
    u[Z] = this->vertices[3 * T(i).vindices[1] + Z] -
           this->vertices[3 * T(i).vindices[0] + Z];

    v[X] = this->vertices[3 * T(i).vindices[2] + X] -
           this->vertices[3 * T(i).vindices[0] + X];
    v[Y] = this->vertices[3 * T(i).vindices[2] + Y] -
           this->vertices[3 * T(i).vindices[0] + Y];
    v[Z] = this->vertices[3 * T(i).vindices[2] + Z] -
           this->vertices[3 * T(i).vindices[0] + Z];

    _glmCross(u, v, &this->facetnorms[3 * (i+1)]);
    _glmNormalize(&this->facetnorms[3 * (i+1)]);
  }
}

/* glmVertexNormals: Generates smooth vertex normals for a model.
 * First builds a list of all the triangles each vertex is in.  Then
 * loops through each vertex in the the list averaging all the facet
 * normals of the triangles each vertex is in.  Finally, sets the
 * normal index in the triangle for the vertex to the generated smooth
 * normal.  If the dot product of a facet normal and the facet normal
 * associated with the first triangle in the list of triangles the
 * current vertex is in is greater than the cosine of the angle
 * parameter to the function, that facet normal is not added into the
 * average normal calculation and the corresponding vertex is given
 * the facet normal.  This tends to preserve hard edges.  The angle to
 * use depends on the model, but 90 degrees is usually a good start.
 *
 * model - initialized GLMmodel structure
 * angle - maximum angle (in degrees) to smooth across
 */
GLvoid CGlmModel::glmVertexNormals(GLfloat angle)
{
  GLMnode*  node;
  GLMnode*  tail;
  GLMnode** members;
//  GLfloat*  normals;
//  GLuint    numnormals;
  GLfloat   average[3];
  GLfloat   dot, cos_angle;
  GLuint    i, avg;

  
  assert(this->facetnorms);

  /* calculate the cosine of the angle (in degrees) */
  cos_angle = float(cos(angle * M_PI / 180.0));

  /* nuke any previous normals */
  if (this->normals)
    free(this->normals);

  /* allocate space for new normals */
  this->numnormals = this->numtriangles * 3; /* 3 normals per triangle */
  this->normals = (GLfloat*)malloc(sizeof(GLfloat)* 3* (this->numnormals+1));

  /* allocate a structure that will hold a linked list of triangle
     indices for each vertex */
  members = (GLMnode**)malloc(sizeof(GLMnode*) * (this->numvertices + 1));
  for (i = 1; i <= this->numvertices; i++)
    members[i] = NULL;
  
  /* for every triangle, create a node for each vertex in it */
  for (i = 0; i < this->numtriangles; i++) {
    node = (GLMnode*)malloc(sizeof(GLMnode));
    node->index = i;
    node->next  = members[T(i).vindices[0]];
    members[T(i).vindices[0]] = node;

    node = (GLMnode*)malloc(sizeof(GLMnode));
    node->index = i;
    node->next  = members[T(i).vindices[1]];
    members[T(i).vindices[1]] = node;

    node = (GLMnode*)malloc(sizeof(GLMnode));
    node->index = i;
    node->next  = members[T(i).vindices[2]];
    members[T(i).vindices[2]] = node;
  }

  /* calculate the average normal for each vertex */
  numnormals = 1;
  for (i = 1; i <= this->numvertices; i++) {
    /* calculate an average normal for this vertex by averaging the
       facet normal of every triangle this vertex is in */
    node = members[i];
    if (!node)
      TRACE("glmVertexNormals(): vertex w/o a triangle\n");
    average[0] = 0.0; average[1] = 0.0; average[2] = 0.0;
    avg = 0;
    while (node) {
      /* only average if the dot product of the angle between the two
         facet normals is greater than the cosine of the threshold
         angle -- or, said another way, the angle between the two
         facet normals is less than (or equal to) the threshold angle */
      dot = _glmDot(&this->facetnorms[3 * T(node->index).findex],
 		    &this->facetnorms[3 * T(members[i]->index).findex]);
      if (dot > cos_angle) {
	node->averaged = GL_TRUE;
	average[0] += this->facetnorms[3 * T(node->index).findex + 0];
	average[1] += this->facetnorms[3 * T(node->index).findex + 1];
	average[2] += this->facetnorms[3 * T(node->index).findex + 2];
	avg = 1;			/* we averaged at least one normal! */
      } else {
	node->averaged = GL_FALSE;
      }
      node = node->next;
    }

    if (avg) {
      /* normalize the averaged normal */
      _glmNormalize(average);

      /* add the normal to the vertex normals list */
      this->normals[3 * numnormals + 0] = average[0];
      this->normals[3 * numnormals + 1] = average[1];
      this->normals[3 * numnormals + 2] = average[2];
      avg = numnormals;
      numnormals++;
    }

    /* set the normal of this vertex in each triangle it is in */
    node = members[i];
    while (node) {
      if (node->averaged) {
	/* if this node was averaged, use the average normal */
	if (T(node->index).vindices[0] == i)
	  T(node->index).nindices[0] = avg;
	else if (T(node->index).vindices[1] == i)
	  T(node->index).nindices[1] = avg;
	else if (T(node->index).vindices[2] == i)
	  T(node->index).nindices[2] = avg;
      } else {
	/* if this node wasn't averaged, use the facet normal */
	this->normals[3 * numnormals + 0] = 
	  this->facetnorms[3 * T(node->index).findex + 0];
	this->normals[3 * numnormals + 1] = 
	  this->facetnorms[3 * T(node->index).findex + 1];
	this->normals[3 * numnormals + 2] = 
	  this->facetnorms[3 * T(node->index).findex + 2];
	if (T(node->index).vindices[0] == i)
	  T(node->index).nindices[0] = numnormals;
	else if (T(node->index).vindices[1] == i)
	  T(node->index).nindices[1] = numnormals;
	else if (T(node->index).vindices[2] == i)
	  T(node->index).nindices[2] = numnormals;
	numnormals++;
      }
      node = node->next;
    }
  }
  
  this->numnormals = numnormals - 1;

  /* free the member information */
  for (i = 1; i <= this->numvertices; i++) {
    node = members[i];
    while (node) {
      tail = node;
      node = node->next;
      free(tail);
    }
  }
  free(members);

  /* pack the normals array (we previously allocated the maximum
     number of normals that could possibly be created (numtriangles *
     3), so get rid of some of them (usually alot unless none of the
     facet normals were averaged)) */
/*  normals = this->normals;
  this->normals = (GLfloat*)malloc(sizeof(GLfloat)* 3* (this->numnormals+1));
  for (i = 1; i <= this->numnormals; i++) {
    this->normals[3 * i + 0] = normals[3 * i + 0];
    this->normals[3 * i + 1] = normals[3 * i + 1];
    this->normals[3 * i + 2] = normals[3 * i + 2];
  }
  free(normals);*/

  printf("glmVertexNormals(): %d normals generated\n", this->numnormals);
}


/* glmLinearTexture: Generates texture coordinates according to a
 * linear projection of the texture map.  It generates these by
 * linearly mapping the vertices onto a square.
 *
 * model - pointer to initialized GLMmodel structure
 */
GLvoid CGlmModel::glmLinearTexture()
{
  GLMgroup *group;
  GLfloat dimensions[3];
  GLfloat x, y, scalefactor;
  GLuint i;
  
  

  if (this->texcoords)
    free(this->texcoords);
  this->numtexcoords = this->numvertices;
  this->texcoords=(GLfloat*)malloc(sizeof(GLfloat)*2*(this->numtexcoords+1));
  
  glmDimensions(dimensions);
  scalefactor = 2.0f / 
    _glmAbs(_glmMax(_glmMax(dimensions[0], dimensions[1]), dimensions[2]));

  /* do the calculations */
  for(i = 1; i <= this->numvertices; i++) {
    x = this->vertices[3 * i + 0] * scalefactor;
    y = this->vertices[3 * i + 2] * scalefactor;
    this->texcoords[2 * i + 0] = (x + 1.0f) / 2.0f;
    this->texcoords[2 * i + 1] = (y + 1.0f) / 2.0f;
  }
  
  /* go through and put texture coordinate indices in all the triangles */
  group = this->groups;
  while(group) {
    for(i = 0; i < group->numtriangles; i++) {
      T(group->triangles[i]).tindices[0] = T(group->triangles[i]).vindices[0];
      T(group->triangles[i]).tindices[1] = T(group->triangles[i]).vindices[1];
      T(group->triangles[i]).tindices[2] = T(group->triangles[i]).vindices[2];
    }    
    group = group->next;
  }

#if 0
  printf("glmLinearTexture(): generated %d linear texture coordinates\n",
	  this->numtexcoords);
#endif
}

/* glmSpheremapTexture: Generates texture coordinates according to a
 * spherical projection of the texture map.  Sometimes referred to as
 * spheremap, or reflection map texture coordinates.  It generates
 * these by using the normal to calculate where that vertex would map
 * onto a sphere.  Since it is impossible to map something flat
 * perfectly onto something spherical, there is distortion at the
 * poles.  This particular implementation causes the poles along the X
 * axis to be distorted.
 *
 * model - pointer to initialized GLMmodel structure
 */
GLvoid CGlmModel::glmSpheremapTexture()
{
  GLMgroup* group;
  GLfloat theta, phi, rho, x, y, z, r;
  GLuint i;
  
  
  assert(this->normals);

  if (this->texcoords)
    free(this->texcoords);
  this->numtexcoords = this->numnormals;
  this->texcoords=(GLfloat*)malloc(sizeof(GLfloat)*2*(this->numtexcoords+1));
     
  /* do the calculations */
  for (i = 1; i <= this->numnormals; i++) {
    z = this->normals[3 * i + 0];	/* re-arrange for pole distortion */
    y = this->normals[3 * i + 1];
    x = this->normals[3 * i + 2];
    r = float(sqrt((x * x) + (y * y)));
    rho = float(sqrt((r * r) + (z * z)));
      
    if(r == 0.0) {
	theta = 0.0f;
	phi = 0.0f;
    } else {
      if(z == 0.0)
	phi = 3.14159265f / 2.0f;
      else
	phi = float(acos(z / rho));
      
#if WE_DONT_NEED_THIS_CODE
      if(x == 0.0)
	theta = 3.14159265f / 2.0f;	/* asin(y / r); */
      else
	theta = acos(x / r);
#endif
      
      if(y == 0.0)
	theta = 3.141592365f / 2.0f;	/* acos(x / r); */
      else
	theta = float(asin(y / r)) + (3.14159265f / 2.0f);
    }
    
    this->texcoords[2 * i + 0] = theta / 3.14159265f;
    this->texcoords[2 * i + 1] = phi / 3.14159265f;
  }
  
  /* go through and put texcoord indices in all the triangles */
  group = this->groups;
  while(group) {
    for (i = 0; i < group->numtriangles; i++) {
      T(group->triangles[i]).tindices[0] = T(group->triangles[i]).nindices[0];
      T(group->triangles[i]).tindices[1] = T(group->triangles[i]).nindices[1];
      T(group->triangles[i]).tindices[2] = T(group->triangles[i]).nindices[2];
    }
    group = group->next;
  }
}

/* glmReadOBJ: Reads a model description from a Wavefront .OBJ file.
 * Returns a pointer to the created object which should be free'd with
 * glmDelete().
 *
 * filename - name of the file containing the Wavefront .OBJ format data.  
 */
GLvoid CGlmModel::glmReadOBJ(char* filename)
{
  FILE*     file;

  /* open the file */
  file = fopen(filename, "r");
  if (!file) 
  {
    TRACE("glmReadOBJ() failed: can't open data file \"%s\".\n", filename);
    exit(1);
  }

  this->pathname      = strdup(filename);

  /* make a first pass through the file to get a count of the number of vertices, normals, texcoords & triangles */
  _glmFirstPass(file);

  /* allocate memory */
  this->vertices = (GLfloat*)malloc(sizeof(GLfloat) * 3 * (this->numvertices + 1));
  this->triangles = (GLMtriangle*)malloc(sizeof(GLMtriangle) * this->numtriangles);
  if (this->numnormals)
    this->normals = (GLfloat*)malloc(sizeof(GLfloat) * 3 * (this->numnormals + 1));
  if (this->numtexcoords)
    this->texcoords = (GLfloat*)malloc(sizeof(GLfloat) * 2 * (this->numtexcoords + 1));

  /* rewind to beginning of file and read in the data this pass */
  rewind(file);

  _glmSecondPass(file);

  /* close the file */
  fclose(file);
}

/* glmWriteOBJ: Writes a model description in Wavefront .OBJ format to
 * a file.
 *
 * model    - initialized GLMmodel structure
 * filename - name of the file to write the Wavefront .OBJ format data to
 * mode     - a bitwise or of values describing what is written to the file
 *            GLM_NONE     -  render with only vertices
 *            GLM_FLAT     -  render with facet normals
 *            GLM_SMOOTH   -  render with vertex normals
 *            GLM_TEXTURE  -  render with texture coords
 *            GLM_COLOR    -  render with colors (color material)
 *            GLM_MATERIAL -  render with materials
 *            GLM_COLOR and GLM_MATERIAL should not both be specified.  
 *            GLM_FLAT and GLM_SMOOTH should not both be specified.  
 */
GLvoid CGlmModel::glmWriteOBJ(char* filename, GLuint mode)
{
  GLuint    i;
  FILE*     file;
  GLMgroup* group;

  

  /* do a bit of warning */
  if (mode & GLM_FLAT && !this->facetnorms) {
    TRACE("glmWriteOBJ() warning: flat normal output requested "
	   "with no facet normals defined.\n");
    mode &= ~GLM_FLAT;
  }
  if (mode & GLM_SMOOTH && !this->normals) {
    TRACE("glmWriteOBJ() warning: smooth normal output requested "
	   "with no normals defined.\n");
    mode &= ~GLM_SMOOTH;
  }
  if (mode & GLM_TEXTURE && !this->texcoords) {
    TRACE("glmWriteOBJ() warning: texture coordinate output requested "
	   "with no texture coordinates defined.\n");
    mode &= ~GLM_TEXTURE;
  }
  if (mode & GLM_FLAT && mode & GLM_SMOOTH) {
    TRACE("glmWriteOBJ() warning: flat normal output requested "
	   "and smooth normal output requested (using smooth).\n");
    mode &= ~GLM_FLAT;
  }

  /* open the file */
  file = fopen(filename, "w");
  if (!file) {
    TRACE("glmWriteOBJ() failed: can't open file \"%s\" to write.\n", filename);
    exit(1);
  }

  /* spit out a header */
  fprintf(file, "#  \n");
  fprintf(file, "#  Wavefront OBJ generated by GLM library\n");
  fprintf(file, "#  \n");
  fprintf(file, "#  GLM library copyright (C) 1997 by Nate Robins\n");
  fprintf(file, "#  email: ndr@pobox.com\n");
  fprintf(file, "#  www:   http://www.pobox.com/~ndr\n");
  fprintf(file, "#  \n");

  if (mode & GLM_MATERIAL && this->mtllibname) {
    fprintf(file, "\nmtllib %s\n\n", this->mtllibname);
    _glmWriteMTL(filename, this->mtllibname);
  }

  /* spit out the vertices */
  fprintf(file, "\n");
  fprintf(file, "# %d vertices\n", this->numvertices);
  for (i = 1; i <= this->numvertices; i++) {
    fprintf(file, "v %f %f %f\n", 
	    this->vertices[3 * i + 0],
	    this->vertices[3 * i + 1],
	    this->vertices[3 * i + 2]);
  }

  /* spit out the smooth/flat normals */
  if (mode & GLM_SMOOTH) {
    fprintf(file, "\n");
    fprintf(file, "# %d normals\n", this->numnormals);
    for (i = 1; i <= this->numnormals; i++) {
      fprintf(file, "vn %f %f %f\n", 
	      this->normals[3 * i + 0],
	      this->normals[3 * i + 1],
	      this->normals[3 * i + 2]);
    }
  } else if (mode & GLM_FLAT) {
    fprintf(file, "\n");
    fprintf(file, "# %d normals\n", this->numfacetnorms);
    for (i = 1; i <= this->numnormals; i++) {
      fprintf(file, "vn %f %f %f\n", 
	      this->facetnorms[3 * i + 0],
	      this->facetnorms[3 * i + 1],
	      this->facetnorms[3 * i + 2]);
    }
  }

  /* spit out the texture coordinates */
  if (mode & GLM_TEXTURE) {
    fprintf(file, "\n");
    fprintf(file, "# %d texcoords\n", this->texcoords);
    for (i = 1; i <= this->numtexcoords; i++) {
      fprintf(file, "vt %f %f\n", 
	      this->texcoords[2 * i + 0],
	      this->texcoords[2 * i + 1]);
    }
  }

  fprintf(file, "\n");
  fprintf(file, "# %d groups\n", this->numgroups);
  fprintf(file, "# %d faces (triangles)\n", this->numtriangles);
  fprintf(file, "\n");

  group = this->groups;
  while(group) {
    fprintf(file, "g %s\n", group->name);
    if (mode & GLM_MATERIAL)
      fprintf(file, "usemtl %s\n", this->materials[group->material].name);
    for (i = 0; i < group->numtriangles; i++) {
      if (mode & GLM_SMOOTH && mode & GLM_TEXTURE) {
	fprintf(file, "f %d/%d/%d %d/%d/%d %d/%d/%d\n",
		T(group->triangles[i]).vindices[0], 
		T(group->triangles[i]).nindices[0], 
		T(group->triangles[i]).tindices[0],
		T(group->triangles[i]).vindices[1],
		T(group->triangles[i]).nindices[1],
		T(group->triangles[i]).tindices[1],
		T(group->triangles[i]).vindices[2],
		T(group->triangles[i]).nindices[2],
		T(group->triangles[i]).tindices[2]);
      } else if (mode & GLM_FLAT && mode & GLM_TEXTURE) {
	fprintf(file, "f %d/%d %d/%d %d/%d\n",
		T(group->triangles[i]).vindices[0],
		T(group->triangles[i]).findex,
		T(group->triangles[i]).vindices[1],
		T(group->triangles[i]).findex,
		T(group->triangles[i]).vindices[2],
		T(group->triangles[i]).findex);
      } else if (mode & GLM_TEXTURE) {
	fprintf(file, "f %d/%d %d/%d %d/%d\n",
		T(group->triangles[i]).vindices[0],
		T(group->triangles[i]).tindices[0],
		T(group->triangles[i]).vindices[1],
		T(group->triangles[i]).tindices[1],
		T(group->triangles[i]).vindices[2],
		T(group->triangles[i]).tindices[2]);
      } else if (mode & GLM_SMOOTH) {
	fprintf(file, "f %d//%d %d//%d %d//%d\n",
		T(group->triangles[i]).vindices[0],
		T(group->triangles[i]).nindices[0],
		T(group->triangles[i]).vindices[1],
		T(group->triangles[i]).nindices[1],
		T(group->triangles[i]).vindices[2], 
		T(group->triangles[i]).nindices[2]);
      } else if (mode & GLM_FLAT) {
	fprintf(file, "f %d//%d %d//%d %d//%d\n",
		T(group->triangles[i]).vindices[0], 
		T(group->triangles[i]).findex,
		T(group->triangles[i]).vindices[1],
		T(group->triangles[i]).findex,
		T(group->triangles[i]).vindices[2],
		T(group->triangles[i]).findex);
      } else {
	fprintf(file, "f %d %d %d\n",
		T(group->triangles[i]).vindices[0],
		T(group->triangles[i]).vindices[1],
		T(group->triangles[i]).vindices[2]);
      }
    }
    fprintf(file, "\n");
    group = group->next;
  }

  fclose(file);
}

/* glmDraw: Renders the model to the current OpenGL context using the
 * mode specified.
 *
 * model    - initialized GLMmodel structure
 * mode     - a bitwise OR of values describing what is to be rendered.
 *            GLM_NONE     -  render with only vertices
 *            GLM_FLAT     -  render with facet normals
 *            GLM_SMOOTH   -  render with vertex normals
 *            GLM_TEXTURE  -  render with texture coords
 *            GLM_COLOR    -  render with colors (color material)
 *            GLM_MATERIAL -  render with materials
 *            GLM_COLOR and GLM_MATERIAL should not both be specified.  
 *            GLM_FLAT and GLM_SMOOTH should not both be specified.  
 */
GLvoid CGlmModel::glmDraw(GLuint mode)
{
  GLuint i;
  GLMgroup* group;

  
  assert(this->vertices);

  /* do a bit of warning */
  if (mode & GLM_FLAT && !this->facetnorms)
  {
    printf("glmDraw() warning: flat render mode requested with no facet normals defined.\n");
    mode &= ~GLM_FLAT;
  }
  if (mode & GLM_SMOOTH && !this->normals)
  {
    printf("glmDraw() warning: smooth render mode requested with no normals defined.\n");
    mode &= ~GLM_SMOOTH;
  }
  if (mode & GLM_TEXTURE && !this->texcoords)
  {
    printf("glmDraw() warning: texture render mode requested with no texture coordinates defined.\n");
    mode &= ~GLM_TEXTURE;
  }
  if (mode & GLM_FLAT && mode & GLM_SMOOTH)
  {
    printf("glmDraw() warning: flat render mode requested and smooth render mode requested (using smooth).\n");
    mode &= ~GLM_FLAT;
  }
  if (mode & GLM_COLOR && !this->materials)
  {
    printf("glmDraw() warning: color render mode requested with no materials defined.\n");
    mode &= ~GLM_COLOR;
  }
  if (mode & GLM_MATERIAL && !this->materials)
  {
    printf("glmDraw() warning: material render mode requested with no materials defined.\n");
    mode &= ~GLM_MATERIAL;
  }
  if (mode & GLM_COLOR && mode & GLM_MATERIAL)
  {
    printf("glmDraw() warning: color and material render mode requested using only material mode\n");
    mode &= ~GLM_COLOR;
  }
  if (mode & GLM_COLOR)
	glEnable(GL_COLOR_MATERIAL);
  if (mode & GLM_MATERIAL)
	glDisable(GL_COLOR_MATERIAL);

  glPushMatrix();
  glTranslatef(this->position[0], this->position[1], this->position[2]);

  glBegin(GL_TRIANGLES);
  group = this->groups;
  while (group)
  {
    if (mode & GLM_MATERIAL)
	{
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, this->materials[group->material].ambient);
      glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, this->materials[group->material].diffuse);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, this->materials[group->material].specular);
      glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, this->materials[group->material].shininess);
    }

    if (mode & GLM_COLOR)
      glColor3fv(this->materials[group->material].diffuse);

    for (i = 0; i < group->numtriangles; i++)
	{
      if (mode & GLM_FLAT)
		glNormal3fv(&this->facetnorms[3 * T(group->triangles[i]).findex]);
      
      if (mode & GLM_SMOOTH)
		glNormal3fv(&this->normals[3 * T(group->triangles[i]).nindices[0]]);
      if (mode & GLM_TEXTURE)
		glTexCoord2fv(&this->texcoords[2*T(group->triangles[i]).tindices[0]]);
      glVertex3fv(&this->vertices[3 * T(group->triangles[i]).vindices[0]]);
      
      if (mode & GLM_SMOOTH)
		glNormal3fv(&this->normals[3 * T(group->triangles[i]).nindices[1]]);
      if (mode & GLM_TEXTURE)
		glTexCoord2fv(&this->texcoords[2*T(group->triangles[i]).tindices[1]]);
      glVertex3fv(&this->vertices[3 * T(group->triangles[i]).vindices[1]]);
     
      if (mode & GLM_SMOOTH)
		glNormal3fv(&this->normals[3 * T(group->triangles[i]).nindices[2]]);
      if (mode & GLM_TEXTURE)
		glTexCoord2fv(&this->texcoords[2*T(group->triangles[i]).tindices[2]]);
      glVertex3fv(&this->vertices[3 * T(group->triangles[i]).vindices[2]]);
    }
    
    group = group->next;
  }
  glEnd();

  glPopMatrix();
}

/* glmList: Generates and returns a display list for the model using
 * the mode specified.
 *
 * model    - initialized GLMmodel structure
 * mode     - a bitwise OR of values describing what is to be rendered.
 *            GLM_NONE     -  render with only vertices
 *            GLM_FLAT     -  render with facet normals
 *            GLM_SMOOTH   -  render with vertex normals
 *            GLM_TEXTURE  -  render with texture coords
 *            GLM_COLOR    -  render with colors (color material)
 *            GLM_MATERIAL -  render with materials
 *            GLM_COLOR and GLM_MATERIAL should not both be specified.  
 *            GLM_FLAT and GLM_SMOOTH should not both be specified.  
 */
GLuint CGlmModel::glmList(GLuint mode)
{
  GLuint list;

  list = glGenLists(1);
  glNewList(list, GL_COMPILE);
  glmDraw(mode);
  glEndList();

  return list;
}

/* glmWeld: eliminate (weld) vectors that are within an epsilon of
 * each other.
 *
 * model      - initialized GLMmodel structure
 * epsilon    - maximum difference between vertices
 *              ( 0.00001 is a good start for a unitized model)
 *
 */
GLvoid CGlmModel::glmWeld(GLfloat epsilon)
{
  GLfloat* vectors;
  GLfloat* copies;
  GLuint   numvectors;
  GLuint   i;

  /* vertices */
  numvectors = this->numvertices;
  vectors    = this->vertices;
  copies = _glmWeldVectors(vectors, &numvectors, epsilon);

  printf("glmWeld(): %d redundant vertices.\n", 
	 this->numvertices - numvectors - 1);

  for (i = 0; i < this->numtriangles; i++) {
    T(i).vindices[0] = (GLuint)vectors[3 * T(i).vindices[0] + 0];
    T(i).vindices[1] = (GLuint)vectors[3 * T(i).vindices[1] + 0];
    T(i).vindices[2] = (GLuint)vectors[3 * T(i).vindices[2] + 0];
  }

  /* free space for old vertices */
  free(vectors);

  /* allocate space for the new vertices */
  this->numvertices = numvectors;
  this->vertices = (GLfloat*)malloc(sizeof(GLfloat) * 
				     3 * (this->numvertices + 1));

  /* copy the optimized vertices into the actual vertex list */
  for (i = 1; i <= this->numvertices; i++) {
    this->vertices[3 * i + 0] = copies[3 * i + 0];
    this->vertices[3 * i + 1] = copies[3 * i + 1];
    this->vertices[3 * i + 2] = copies[3 * i + 2];
  }

  free(copies);
}


#if 0
  /* normals */
  if (this->numnormals) {
  numvectors = this->numnormals;
  vectors    = this->normals;
  copies = _glmOptimizeVectors(vectors, &numvectors);

  printf("glmOptimize(): %d redundant normals.\n", 
	 this->numnormals - numvectors);

  for (i = 0; i < this->numtriangles; i++) {
    T(i).nindices[0] = (GLuint)vectors[3 * T(i).nindices[0] + 0];
    T(i).nindices[1] = (GLuint)vectors[3 * T(i).nindices[1] + 0];
    T(i).nindices[2] = (GLuint)vectors[3 * T(i).nindices[2] + 0];
  }

  /* free space for old normals */
  free(vectors);

  /* allocate space for the new normals */
  this->numnormals = numvectors;
  this->normals = (GLfloat*)malloc(sizeof(GLfloat) * 
				    3 * (this->numnormals + 1));

  /* copy the optimized vertices into the actual vertex list */
  for (i = 1; i <= this->numnormals; i++) {
    this->normals[3 * i + 0] = copies[3 * i + 0];
    this->normals[3 * i + 1] = copies[3 * i + 1];
    this->normals[3 * i + 2] = copies[3 * i + 2];
  }

  free(copies);
  }

  /* texcoords */
  if (this->numtexcoords) {
  numvectors = this->numtexcoords;
  vectors    = this->texcoords;
  copies = _glmOptimizeVectors(vectors, &numvectors);

  printf("glmOptimize(): %d redundant texcoords.\n", 
	 this->numtexcoords - numvectors);

  for (i = 0; i < this->numtriangles; i++) {
    for (j = 0; j < 3; j++) {
      T(i).tindices[j] = (GLuint)vectors[3 * T(i).tindices[j] + 0];
    }
  }

  /* free space for old texcoords */
  free(vectors);

  /* allocate space for the new texcoords */
  this->numtexcoords = numvectors;
  this->texcoords = (GLfloat*)malloc(sizeof(GLfloat) * 
				      2 * (this->numtexcoords + 1));

  /* copy the optimized vertices into the actual vertex list */
  for (i = 1; i <= this->numtexcoords; i++) {
    this->texcoords[2 * i + 0] = copies[2 * i + 0];
    this->texcoords[2 * i + 1] = copies[2 * i + 1];
  }

  free(copies);
  }
#endif

#if 0
  /* look for unused vertices */
  /* look for unused normals */
  /* look for unused texcoords */
  for (i = 1; i <= this->numvertices; i++) {
    for (j = 0; j < this->numtriangles; i++) {
      if (T(j).vindices[0] == i || 
	  T(j).vindices[1] == i || 
	  T(j).vindices[1] == i)
	break;
    }
  }
#endif
